package fsmv;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import fsmv.phi.Configuration;

public final class ComposedFSMv extends FSMv 
{
	List<FSMv> fsmvs;
	FSMvRho compositionConstraint;
	
	public ComposedFSMv(String selectedDirPath, String line)
	{
		
		String[] elements=line.trim().split(":");
	
		if(elements.length!=3)
		{
			throw new RuntimeException("");
		}
		
		this.name=elements[0];
		
		elements[1]=elements[1].trim();
		if(elements[1].startsWith("(") && elements[1].endsWith(")"))
		{
			elements[1]=elements[1].substring(1, elements[1].length()-1);
		}
		
		
		if(elements[1].length()==0)
			throw new RuntimeException("A composedFSMv should be composed of at least one FSMv. Here listOfFSMvs.size()=0");
		
		
		String[] machines=elements[1].split(",");
		this.fsmvs=new ArrayList<>();
		
		for(String machine: machines)
		{
			if(!machine.contains(":"))
				this.fsmvs.add(new SingleFSMv(selectedDirPath+"/"+machine));
			else
				this.fsmvs.add(new ComposedFSMv(selectedDirPath, machine));
		}
				
		this.type=fsmvs.get(0).getType();
		for(FSMv fsmv:fsmvs)
		{
			if(fsmv.getType()!=this.type)
			{
				throw new RuntimeException("The composed FSMv under creation is mixing requirement machine with design machine");
			}
		}
		
		
		if(!elements[2].toLowerCase().endsWith(".tvl"))
		{
			throw new RuntimeException("The composition constraint should be a TVL file");
		}
		
		this.compositionConstraint=new FSMvRho(new File(selectedDirPath+"/"+elements[2]), type=='D');
		
	}
	
        public ComposedFSMv(List<FSMv> fsmvs, FSMvRho compositionConstraint)
        {
            this.fsmvs=fsmvs;
            this.compositionConstraint=compositionConstraint;
        }
        
	@Override
	public void generateValidConfiguration() 
	{
            this.listOfValidConfiguration=new ArrayList<>();
            List<Configuration> localListOfValidConfiguration=fsmvs.get(0).getListOfValidConfiguration();
            for(int i=1;i<fsmvs.size();i++)
            {
                    localListOfValidConfiguration=mergeWithConfigurations(localListOfValidConfiguration, fsmvs.get(i));
            }

            for(Configuration conf: localListOfValidConfiguration)
            {
                    if(this.compositionConstraint.isSatisfiedBy(conf))
                            this.listOfValidConfiguration.add(conf);
            }
	}
	

	private List<Configuration> mergeWithConfigurations(List<Configuration> existingConfiguration, FSMv fsmv) 
	{
		List<Configuration> result=new ArrayList<>(existingConfiguration.size()*fsmv.getListOfValidConfiguration().size());
		for(Configuration newConfiguration: fsmv.getListOfValidConfiguration())
		{
			for(Configuration mergedConfiguration: existingConfiguration)
			{
				Configuration conf=new Configuration();
				conf.getListOfVariable_value().putAll(mergedConfiguration.getListOfVariable_value());
				conf.getListOfVariable_value().putAll(newConfiguration.getListOfVariable_value());
				result.add(conf);
			}
		}
		return result;
	}

	@Override
	public List<FSMvVariable> getVariables() 
	{
		List<FSMvVariable> result=new ArrayList<>();
		
		for(FSMv fsmv: this.fsmvs)
		{
			result.addAll(fsmv.getVariables());
		}
		
		return result;
	}
         
         
	@Override
	public FSMvRho getRho() 
	{
		List<FSMvRho> rhos=new ArrayList<>();
		
		rhos.add(compositionConstraint);
		
		for(FSMv fsmv: fsmvs)
			rhos.add(fsmv.getRho());
		
		return FSMvRho.merge(rhos);
	}

	@Override
	public FSMvVariable getVariableFromName(String varName) 
	{
		FSMvVariable result=null;
		for(FSMv fsmv: fsmvs)
		{
			result=fsmv.getVariableFromName(varName);
			if(result!=null)
				return result;
		}
		return null;
	}

	@Override
	public String toString() 
	{
		StringBuilder sb=new StringBuilder();
		for(FSMv fsmv: fsmvs)
		{
			sb.append(fsmv.toString());
			sb.append("\n\n");
		}
		sb.append(this.compositionConstraint.toString()).append("\n");
		
		return sb.toString();
	}

    @Override
    public List<FSMvEvent> getEvents() 
    {
        List<FSMvEvent> result = new ArrayList<>();
        for(FSMv fsmv: fsmvs)
        {
            for(FSMvEvent event: fsmv.getEvents())
            {
                if(!result.contains(event))
                    result.add(event);
            }
        }
        return result;
    }
    
    public int getNbOfSingleFSMv()
    {
        return getSingleFSMvs().size();
    }
    
    public List<SingleFSMv> getSingleFSMvs()
    {
        List<SingleFSMv> result=new ArrayList<>();
        for(FSMv fsmv:this.fsmvs)
        {
            if(fsmv instanceof SingleFSMv)
                result.add((SingleFSMv)fsmv);
            else if(fsmv instanceof ComposedFSMv)
                result.addAll(((ComposedFSMv)fsmv).getSingleFSMvs());
        }
        return result;
    }
}
