package lucci.io.stream;

import java.io.*;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import java.util.zip.*;

import toools.io.Utilities;
import toools.text.TextUtilities;


/**
 * This abstract subclass of StreamSource handles stream access restrictions,
 * compression and buffering.
 * 
 * @author Luc Hogie
 */

public abstract class AbstractStreamSource implements StreamSource
{
	/**
	 * @return if the document source is compressed.
	 */
	@Override
	public boolean isCompressed()
	{
		String name = getName();

		if (name == null)
		{
			return false;
		}
		else
		{
			return name.endsWith(".gz");
		}
	}

	@Override
	public final InputStream createInputStream() throws IOException
	{
		if (isReadable())
		{
			InputStream is = createInputStreamImpl();
			BufferedInputStream bis = new BufferedInputStream(is);// , 1024);

			if (isCompressed())
			{
				return new GZIPInputStream(bis);
			}
			else
			{
				return bis;
			}
		}
		else
		{
			String className = getClass().getName();
			className = className.substring(className.lastIndexOf('.') + 1);
			throw new IOException(getName() + " is not readable");
		}
	}

	@Override
	public final OutputStream createOutputStream() throws IOException
	{
		if (isWritable())
		{
			OutputStream os = createOutputStreamImpl();
			BufferedOutputStream bos = new BufferedOutputStream(os);

			if (isCompressed())
			{
				return new GZIPOutputStream(bos);
			}
			else
			{
				return bos;
			}
		}
		else
		{
			String className = getClass().getName();
			className = className.substring(className.lastIndexOf('.') + 1);
			throw new IOException(getName() + " is not writable");
		}
	}

	/**
	 * @return the input stream for the document.
	 */
	public abstract InputStream createInputStreamImpl() throws IOException;

	/**
	 * @return the output stream for the document.
	 */
	public abstract OutputStream createOutputStreamImpl() throws IOException;

	@Override
	public byte[] readItAll() throws IOException
	{
		return Utilities.readUntilEOF(createInputStream());
	}

	public void writeItAll(byte[] bytes) throws IOException
	{
		if (bytes == null) throw new NullPointerException("null byte array");

		OutputStream os = createOutputStream();
		os.write(bytes);
		os.flush();
		os.close();
	}

	public List<String> getLines() throws IOException
	{
		return Arrays.asList(new String(readItAll()).split("\\n"));
	}

	/**
	 * Sets the source as a string, relatively to the given source.
	 */
	@Override
	public abstract void setAsText(StreamSource reference, String name);
}
