package toools;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import toools.io.JavaResource;


public class Clazz
{
    public static <T> List<Class<T>> findImplementationsInTheSamePackage(Class<T> model)
    {
	return findImplementations(model, Clazz.listAllClasses(model.getPackage()));
    }

    public static <T> List<Class<T>> findImplementationsInPackage(Class<T> model, Package... pkg)
    {
	return findImplementations(model, Clazz.listAllClasses(pkg));
    }

    public static <T> List<Class<T>> findImplementations(Class<T> model)
    {
	return findImplementations(model, ClassPath.retrieveClassPath());
    }

    public static <T> List<Class<T>> findImplementations(Class<T> model, ClassPath classpath)
    {
	return findImplementations(model, classpath.listAllClasses());
    }

    public static <T> List<Class<T>> findImplementations(Class<T> model, Collection<Class<?>> classes)
    {
	List<Class<T>> encoderList = new ArrayList<Class<T>>();

	for (Class<?> potentialImplementation : classes)
	{
	    if (isAnImplementation(potentialImplementation, model))
	    {
		encoderList.add((Class<T>) potentialImplementation);
	    }
	}

	return encoderList;
    }


    
    public static boolean isAnImplementation(Class<?> potentialImplementation, Class<?> model)
    {
	return model.isAssignableFrom(potentialImplementation) && Clazz.isInstantiable(potentialImplementation);
    }

    public static boolean classExists(String className)
    {
	try
	{
	    Class.forName(className);
	    return true;
	}
	catch (ClassNotFoundException e)
	{
	    return false;
	}
	catch (NoClassDefFoundError e)
	{
	    return false;
	}
    }

    public static boolean isAbstract(Class<?> thisClass)
    {
	return Modifier.isAbstract(thisClass.getModifiers());
    }

    public static boolean isInterface(Class<?> thisClass)
    {
	return Modifier.isInterface(thisClass.getModifiers());
    }

    public static boolean isConcrete(Class<?> thisClass)
    {
	return !isInterface(thisClass) && !isAbstract(thisClass);
    }

    public static <V> Constructor<V> findDefaultConstructor(Class<V> thisClass)
    {
	try
	{
	    return thisClass.getConstructor(new Class[0]);
	}
	catch (SecurityException e)
	{
	    return null;
	}
	catch (NoSuchMethodException e)
	{
	    return null;
	}
    }

    public static boolean isInner(Class<?> thisClass)
    {
	return thisClass.getName().contains("$");
    }

    public static boolean isInstantiable(Class<?> thisClass)
    {
	return Modifier.isPublic(thisClass.getModifiers()) && isConcrete(thisClass) && !isInner(thisClass);
    }

    public static boolean isInstantiableWithoutArguments(Class<?> thisClass)
    {
	return isInstantiable(thisClass) && findDefaultConstructor(thisClass) != null;
    }

    public static Class<?> findClass(String className) throws NoClassDefFoundError
    {
	try
	{
	    return Class.forName(className);
	}
	catch (ClassNotFoundException e)
	{
	    throw new NoClassDefFoundError(className);
	}
    }

    public static Class<?> findClassOrFail(String className)
    {
	if (className == null)
	    throw new NullPointerException();

	try
	{
	    return Class.forName(className);
	}
	catch (ClassNotFoundException e)
	{
	    throw new NoClassDefFoundError(e.getMessage());
	}
    }

    public static <T> T makeInstanceOrFail(Class<T> clazz) throws ClassInstantiationException
    {
	if (clazz == null)
	    throw new NullPointerException("cannot instantiate the null class");

	try
	{
	    return clazz.newInstance();
	}
	catch (Throwable e)
	{
	    if (e.getCause() == null)
	    {
		throw new ClassInstantiationException(e);
	    }
	    else
	    {
		throw new ClassInstantiationException(e.getCause());
	    }
	}
    }

    public static <T> T makeInstance(Class<T> clazz)
    {
	try
	{
	    return makeInstanceOrFail(clazz);
	}
	catch (ClassInstantiationException e)
	{
	    throw new IllegalStateException(e);
	}
    }

    public static <T> T makeInstance(Constructor<T> constructor, Object... args)
    {
	try
	{
	    return constructor.newInstance(args);
	}
	catch (Exception e)
	{
	    if (e.getCause() == null)
	    {
		e.printStackTrace();
	    }
	    else
	    {
		e.getCause().printStackTrace();
	    }

	    return null;
	}
    }

    public static <T> List<T> makeSeveralInstances(Class<T> clazz, int numberOfInstances)
	    throws ClassInstantiationException
    {
	List<T> c = new ArrayList<T>();

	while (numberOfInstances-- > 0)
	{
	    c.add(makeInstance(clazz));
	}

	return c;
    }

    public static List<Class<?>> listAllClasses(Package... pkg)
    {
	List<Package> packages = Arrays.asList(pkg);
	List<Class<?>> classes = new ArrayList<Class<?>>();

	for (Class<?> clazz : ClassPath.retrieveClassPath().listAllClasses())
	{
	    if (packages.contains(clazz.getPackage()))
	    {
		classes.add(clazz);
	    }
	}

	return classes;
    }

    public static boolean isExecutable(Class<?> thisClass)
    {
	try
	{
	    Method m = thisClass.getMethod("main", new Class[] { (new String[0]).getClass() });
	    return (m.getModifiers() & Modifier.STATIC) != 0;
	}
	catch (Throwable e)
	{
	    return false;
	}
    }

    public static ClassPath findClassContainer(String classname, ClassPath urls)
    {
	String resname = classname.replace('.', '/') + ".class";
	return JavaResource.findResourceContainer(resname, urls);
    }

    public static boolean hasDefaultConstructor(Class<? extends Object> c)
    {
	try
	{
	    return c.getConstructor(new Class[0]) != null;
	}
	catch (SecurityException e)
	{
	    throw new IllegalStateException();
	}
	catch (NoSuchMethodException e)
	{
	    return false;
	}
    }

}
