package toools.extern;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import toools.ExceptionUtilities;
import toools.collections.Collections;
import toools.io.Utilities;
import toools.io.file.Directory;

public class Proces
{
    public static boolean TRACE_CALLS = false;

    private static class ReadThread extends Thread
    {
	boolean hasCompleted = false;
	private BufferedInputStream is;
	ByteArrayOutputStream targetBuffer = new ByteArrayOutputStream();

	public ReadThread(InputStream is)
	{
	    this.is = new BufferedInputStream(is);
	    start();
	}

	public void run()
	{
	    synchronized (this)
	    {
		try
		{
		    // get all the content from the I/O stream and put
		    // it into a byte array
		    Utilities.copy(is, targetBuffer);

		    // that it's finished, flag it!
		    hasCompleted = true;

		    // resume the thread that was waiting for retrieving the
		    // data
		    notify();
		}
		catch (IOException e)
		{
		    e.printStackTrace();
		}
	    }
	}
    }

    public static byte[] exec(String cmd, Directory directory, byte[] stdin, String... args)
    {
	try
	{
	    ProcessOutput output = rawExec(cmd, directory, stdin, args);

	    // the execution went ok
	    if (output.getReturnCode() == 0)
	    {
		return output.getStdout();
	    }
	    else
	    {
		byte[] out = output.getStdout();
		byte[] err = output.getStderr();
		String error = new String(err.length == 0 ? out : err);
		throw new IllegalStateException("Command " + cmd + " " + Arrays.asList(args)
			+ " has failed (exit code=" + output.getReturnCode() + ")\n" + error);
	    }
	}
	catch (IOException e)
	{
	    throw ExceptionUtilities.toRuntimeException(e);
	}
    }

    public static byte[] exec(String cmd, Directory d, String... args)
    {
	return exec(cmd, d, null, args);
    }

    public static byte[] exec(String cmd, byte[] sdtin, String... args)
    {
	return exec(cmd, Directory.getCurrentDirectory(), sdtin, args);
    }

    public static byte[] exec(String cmd, String... arg)
    {
	return exec(cmd, Directory.getCurrentDirectory(), null, arg);
    }

    public static ProcessOutput rawExec(String cmd, Directory directory, byte[] stdin, String... args)
	    throws IOException
    {
	try
	{
	    List<String> tokens = new ArrayList<String>();
	    tokens.add(cmd);

	    if (args != null)
	    {
		tokens.addAll(Arrays.asList(args));
	    }

	    if (TRACE_CALLS)
		System.out.println("Executing process:\n\t" + Collections.toString(tokens, " ")
			+ "\nrunning in directory:\n\t" + directory.getPath());

	    Process process = Runtime.getRuntime().exec(tokens.toArray(new String[0]), null, directory.toFile());

	    if (stdin != null)
	    {
		process.getOutputStream().write(stdin);
		process.getOutputStream().flush();
		process.getOutputStream().close();
	    }

	    // streams MUST be read in parallel of the process in order
	    // to prevent buffer overflows (which block the process)
	    // System.out.println(1);

	    ReadThread stdoutThread = new ReadThread(process.getInputStream());
	    ReadThread stderrthread = new ReadThread(process.getErrorStream());

	    process.waitFor();

	    for (ReadThread t : new ReadThread[] { stdoutThread, stderrthread })
	    {
		synchronized (t)
		{
		    if (!t.hasCompleted)
		    {
			t.wait();
		    }
		}
	    }

	    return new ProcessOutput(process.exitValue(), stdoutThread.targetBuffer.toByteArray(),
		    stderrthread.targetBuffer.toByteArray());
	}
	catch (InterruptedException e)
	{
	    e.printStackTrace();
	    throw new IllegalStateException("command " + cmd + " has been interrupted");
	}
    }

    public static ProcessOutput rawExec(String cmd, Directory d, String... args) throws IOException
    {
	return rawExec(cmd, d, null, args);
    }

    public static ProcessOutput rawExec(String cmd, byte[] in, String... args) throws IOException
    {
	return rawExec(cmd, Directory.getCurrentDirectory(), in, args);
    }

    public static ProcessOutput rawExec(String cmd, String... args) throws IOException
    {
	return rawExec(cmd, Directory.getCurrentDirectory(), null, args);
    }

}
