package toools.io;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import toools.ClassContainer;
import toools.ClassPath;
import toools.io.file.AbstractFile;
import toools.io.file.Directory;
import toools.io.file.RegularFile;

public class JavaResource
{
    private String name;
    private Class<?> refClass;

    // where is stored the resource
    private ClassContainer container;

    public JavaResource(String name)
    {
	if (!name.startsWith("/"))
	    throw new IllegalArgumentException("resource name should start with a slash: " + name);

	this.name = name;
    }

    public JavaResource(Class<?> clazz, String name)
    {
	if (name.startsWith("/"))
	    throw new IllegalArgumentException(
		    "no slash allowed in resource name since its location is already characterized by a class");

	this.refClass = clazz;
	this.name = '/' + clazz.getPackage().getName().replace('.', '/') + '/' + name;
    }

    public String getPath()
    {
	return name;
    }

    public void setName(String newName)
    {
	if (newName == null)
	    throw new IllegalArgumentException();

	name = newName;
    }

    public InputStream getInputStream()
    {
	if (refClass == null)
	{
	    return getClass().getResourceAsStream(getPath());
	}
	else
	{
	    return refClass.getResourceAsStream(getPath());
	}
    }

    public URL getURL()
    {
	return refClass.getResource(getPath());
    }

    public byte[] getByteArray() throws IOException
    {
	InputStream is = getInputStream();

	if (is == null)
	{
	    throw new IOException("cannot get an input stream for resource " + getPath());
	}
	else
	{
	    return Utilities.readUntilEOF(is);
	}
    }

    public static List<JavaResource> listResources()
    {
	return listResources(ClassPath.retrieveClassPath());
    }

    public static List<JavaResource> listResources(ClassPath urls)
    {
	List<JavaResource> resources = new ArrayList<JavaResource>();

	for (ClassContainer thisClassPathEntry : urls)
	{
	    resources.addAll(listResources(thisClassPathEntry));
	}

	return resources;
    }

    public static List<JavaResource> listResources(ClassContainer classPathEntry)
    {
	List<JavaResource> resources = new ArrayList<JavaResource>();
	AbstractFile entryFile = classPathEntry.getFile();

	if (!entryFile.exists())
	    throw new IllegalStateException("classpath entry not found: " + entryFile.getPath());

	if (entryFile instanceof Directory)
	{
	    for (AbstractFile sf : ((Directory) entryFile).retrieveTree(
		    new toools.io.file.FileFilter.RegularFileFilter(), null))
	    {
		String resname = sf.getPath().replace(entryFile.getPath(), "");
		JavaResource res = new JavaResource(resname);
		res.setContainer(classPathEntry);
		resources.add(res);
	    }
	}
	else if (entryFile.getName().matches(".*\\.(jar|zip)$"))
	{
	    try
	    {
		ZipFile zf = new ZipFile(new File(entryFile.getPath()));
		Enumeration entries = zf.entries();

		while (entries.hasMoreElements())
		{
		    ZipEntry e = (ZipEntry) entries.nextElement();
		    String entryName = e.getName();

		    if (!entryName.startsWith("/"))
			entryName = "/" + entryName;

		    JavaResource thisResource = new JavaResource(entryName);
		    thisResource.setContainer(classPathEntry);
		    resources.add(thisResource);
		}
	    }
	    catch (ZipException e1)
	    {
		System.err.println("File " + entryFile.getPath());
		e1.printStackTrace();
	    }
	    catch (IOException e1)
	    {
		System.err.println("File " + entryFile.getPath());
		e1.printStackTrace();
	    }
	}
	else
	{
	    // throw new IllegalStateException("dunno how to handle " +
	    // thisClassPathEntry.getName());
	    System.err.println("dunno how to handle " + entryFile.getName());
	}

	return resources;
    }

    public ClassContainer getContainer()
    {
	return this.container;
    }

    public void setContainer(ClassContainer container)
    {
	this.container = container;
    }

    public static ClassPath findResourceContainer(String name, ClassPath classpath)
    {
	ClassPath urls = new ClassPath();

	// for all the resources found in the classpath
	for (JavaResource r : JavaResource.listResources(classpath))
	{
	    // if this resource has the same name as the resource looked for
	    if (r.getPath().equals(name))
	    {
		urls.add(r.getContainer());
	    }
	}

	return urls;
    }

    public boolean exists()
    {
	InputStream is = getInputStream();

	if (is == null)
	{
	    return false;
	}
	else
	{
	    try
	    {
		is.close();
	    }
	    catch (IOException e)
	    {
	    }

	    return true;
	}
    }

    public void exportToFile(RegularFile f) throws IOException
    {
	if (!f.getParent().exists())
	    f.getParent().mkdirs();

	f.setContent(getByteArray());
    }

    public static void exportToFile(String resourceName, RegularFile dest) throws IOException
    {
	if (!dest.getParent().exists())
	    dest.getParent().mkdirs();

	new JavaResource(resourceName).exportToFile(dest);
    }

    public static void loadLibrary(String resourceName) throws IOException
    {
	RegularFile f = new RegularFile(Directory.getSystemTempDirectory(), resourceName);
	exportToFile(resourceName, f);
	System.load(f.getPath());
    }

    public String getName()
    {
	return name.substring(name.lastIndexOf('/') + 1);
    }

    public static JavaResource getAttachedResource(Class<?> referenceClass, String resourceName)
    {
	return new JavaResource('/' + referenceClass.getPackage().getName().replace('.', '/') + '/' + resourceName);
    }
    
}
