package toools.io;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;
import java.io.OutputStream;
import java.util.Map;
import java.util.Properties;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import toools.CodeShouldNotHaveBeenReachedException;
import toools.collections.Maps;

public class Utilities
{
    public static boolean operatingSystemIsUNIX()
    {
	return new File("/etc/passwd").exists();
    }

    public static Map<String, String> loadPropertiesToMap(String text)
    {
	try
	{
	    Properties properties = new Properties();
	    properties.load(new ByteArrayInputStream(text.getBytes()));
	    return Maps.propertiesToMap(properties);
	}
	catch (IOException e)
	{
	    throw new CodeShouldNotHaveBeenReachedException();
	}
    }

    public static byte[] gzip(byte[] data)
    {
	try
	{
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    GZIPOutputStream gos = new GZIPOutputStream(bos);
	    gos.write(data);
	    gos.close();
	    return bos.toByteArray();
	}
	catch (IOException ex)
	{
	    throw new IllegalStateException();
	}
    }

    public static byte[] gunzip(byte[] data)
    {
	try
	{
	    ByteArrayInputStream bis = new ByteArrayInputStream(data);
	    GZIPInputStream gis = new GZIPInputStream(bis);
	    byte[] uncompressedData = readUntilEOF(gis);
	    return uncompressedData;
	}
	catch (IOException ex)
	{
	    throw new IllegalStateException();
	}
    }

    public static byte[] readUntilEOF(InputStream is) throws IOException
    {
	if (is == null)
	    throw new NullPointerException("null stream");

	ByteArrayOutputStream bos = new ByteArrayOutputStream();
	copy(is, bos);
	return bos.toByteArray();
    }

    public static void copy(InputStream is, OutputStream bos) throws IOException
    {
	if (is == null)
	    throw new NullPointerException();

	if (bos == null)
	    throw new NullPointerException();

	byte[] bytes = new byte[4 * 1024];

	while (true)
	{
	    int i = is.read(bytes);

	    if (i == -1)
	    {
		break;
	    }
	    else
	    {
		bos.write(bytes, 0, i);
	    }
	}
    }

    public static byte[] objectToBytes(Object o)
    {
	try
	{
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    ObjectOutputStream oos = new ObjectOutputStream(bos);
	    oos.writeObject(o);
	    oos.flush();
	    return bos.toByteArray();
	}
	catch (IOException ex)
	{
	    ex.printStackTrace();
	    throw new CodeShouldNotHaveBeenReachedException();
	}
    }

    public static class CustomObjectInputStream extends ObjectInputStream
    {
	private ClassLoader classLoader;

	public CustomObjectInputStream(InputStream in, ClassLoader classLoader) throws IOException
	{
	    super(in);
	    this.classLoader = classLoader;
	}

	protected Class<?> resolveClass(ObjectStreamClass desc) throws IOException, ClassNotFoundException
	{
	    // System.out.println("using classloader " +
	    // classLoader.getClass().getName() + " load class " +
	    // desc.getName());
	    return Class.forName(desc.getName(), false, classLoader);
	}
    }

    public static <T> T bytesToObject(Class<T> clazz, byte[] bytes)
    {
	return bytesToObject(clazz, bytes, Thread.currentThread().getContextClassLoader());
    }

    public static <T> T bytesToObject(Class<T> clazz, byte[] bytes, ClassLoader classLoader)
    {
	try
	{
	    return (T) new CustomObjectInputStream(new ByteArrayInputStream(bytes), classLoader).readObject();
	}
	catch (IOException e)
	{
	    IllegalStateException e2 = new IllegalStateException();
	    e2.setStackTrace(e.getStackTrace());
	    throw e2;
	}
	catch (ClassNotFoundException e)
	{
	    throw new IllegalStateException("cannot find class " + e.getMessage());
	}
    }

    public static void pressEnterToContinue()
    {
	Utilities.readUserInput("Press enter to continue", ".*");
    }

    public static void pressEnterToContinue(String msg)
    {
	Utilities.readUserInput(msg, ".*");
    }

    
    public static String readUserInput(String invitation, String regexp)
    {
	try
	{
	    BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));

	    while (true)
	    {
		if (invitation != null)
		{
		    System.out.print(invitation);
		}

		String line = stdin.readLine();

		if (regexp == null || line == null || line.matches(regexp))
		{
		    return line;
		}
		else
		{
		    // System.err.println("Input error, the string must match "
		    // + regexp + "\n");
		}
	    }
	}
	catch (IOException e)
	{
	    System.err.println("Error while reading user input");
	    return null;
	}
    }


}
