package toools.io.data_transfer;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.List;
import java.util.Vector;




public class DataTransfer
{
	/* the input stream the data is read from */
	private InputStream inputStream;

	/* the output stream the data is written to */
	private OutputStream outputStream;

	/* the length of the data, -1 if unknown */
	private int expectedDataLength;

	/* the transfer notification step (in bytes) */
	private int notificationStep;

	/* the transfer notification delay (in milliseconds) */
	private long notificationDelay;

	/* the amount of transfered data (in bytes) */
	private int transferedDataLength;

	/* the begin date of the transfer */
	private long beginDate;

	/* the end date of the transfer */
	private long endDate;

	/* the attempt count when errors occur */
	private int attemptCount;

	/* the transfer rate */
	private int transferRate;

	/* the listeners */
	private List<DataTransferListener> listeners;

	/* the thread */
	private Thread thread;


	public DataTransfer()
	{
		// the stream are stdin and stdout by default
		inputStream = System.in;
		outputStream = System.out;

		// the data length is unknown
		expectedDataLength = -1;

		// the listeners are not notified of packet arrival
		notificationStep = -1;

		// the listeners are never notified
		notificationDelay = 1000;

		// no data has been transfered
		transferedDataLength = 0;

		// the transfer has not yet begun, and so cannot be terminated
		beginDate = -1;
		endDate = -1;

		// on error, always retry once
		attemptCount = 1;

		// the transfer rate is undefined
		transferRate = -1;

		// there's no listeners
		listeners = null;

		// no thread is required
		thread = null;
	}
	public void addDataTransferListener( DataTransferListener l )
	{
		if ( listeners == null )
		{
			listeners = new Vector<DataTransferListener>();
		}

		listeners.add( l );
	}
	/**
	 * @return the number of re-attempts to do when an error occur.
	 */
	public int getAttemptCount()
	{
		return attemptCount;
	}
	/**
	 * @return the average transfer rate (byte/s). This is the average
	 * transfer rate of the whole transfer.
	 */
	public int getAverageTransferRate()
	{
		if ( beginDate == -1 )
			throw new IllegalStateException( "transfer has not yet begun" );

		long date = endDate == -1 ? System.currentTimeMillis() : endDate;
		long delay = date - beginDate;
		return (int) ((1000L * transferedDataLength) / (delay == 0 ? 1 : delay));
	}
	/**
	 * @return the begin date of the transfer.
	 */
	public long getBeginDate()
	{
		if ( beginDate == -1 )
			throw new IllegalStateException( "transfer has not yet begun" );

		return beginDate;
	}
	/**
	 * @return the end date of the transfer.
	 */
	public long getEndDate()
	{
		if ( endDate == -1 )
			throw new IllegalStateException( "transfer is not terminated" );

		return endDate;
	}
	/**
	 * @return the expected length of the data (in bytes) to transfer or
	 * -1 if the length is unknown.
	 */
	public int getExpectedDataLength()
	{
		return expectedDataLength;
	}
	/**
	 * @return the input stream the data is read from.
	 */
	public InputStream getInputStream()
	{
		return inputStream;
	}
	/**
	 * @return the notification delay (in milliseconds).
	 */
	public long getNotificationDelay()
	{
		return notificationDelay;
	}
	/**
	 * @return the notification step (in bytes).
	 */
	public int getNotificationStep()
	{
		return notificationStep;
	}
	/**
	 * @return the output stream the data is written to.
	 */
	public OutputStream getOutputStream()
	{
		return outputStream;
	}
	/**
	 * @return the remaining delay.
	 */
	public long getRemainingDelay()
	{
		if ( beginDate == -1 )
			throw new IllegalStateException();

		if ( transferedDataLength < 0 )
			throw new IllegalStateException();

		if ( transferedDataLength == 0 )
		{
			return -1;
		}
		else
		{
			long date = endDate == -1 ? System.currentTimeMillis() : endDate;
			long delay = date - beginDate;
			return ((expectedDataLength - transferedDataLength) * delay) / transferedDataLength;
		}
	}
	/**
	 * @return the total amount of transfered bytes.
	 */
	public int getTransferedDataLength()
	{
		return transferedDataLength;
	}
	/**
	 * @return the instant transfer rate (byte/s), -1 if undefined.
	 * This 'instant' rate is n fact the average transfer rate between
	 * the 2 last periodic notifications.
	 */
	public int getTransferRate()
	{
		return transferRate == -1 ? getAverageTransferRate() : transferRate;
	}
	/**
	 * @return the ratio of data that have been received.
	 * A ratio of 0 means no data has been received, a ratio
	 * of 1 means the data has been fully read.
	 */
	public double getTransferRatio()
	{
		// if the expected length is undefined, the ratio has no sense
		if ( expectedDataLength == -1 )
			throw new IllegalStateException( "cannot evaluate" );

		return ((double) transferedDataLength) / ((double) expectedDataLength);
	}
	/**
	 * @return if the transfer is stalled. It's stalled if the instantaneous
	 * tranfer rate is 0 bytes/s.
	 */
	public boolean isStalled()
	{
		return getTransferRate() == 0;
	}
	/**
	 * Transfer the data.
	 */
	public void process()
		throws IOException
	{
		// if a notification delay is defined, a thread must be created
		if ( notificationDelay != -1 )
		{
			thread = new Thread();
			thread.start();
		}

		// store the begin date of the transfer
		beginDate = System.currentTimeMillis();

		processTransferStarted( new DataTransferEvent( this ) );

		try
		{
			processImpl();
		}
		catch ( IOException ex )
		{
			throw ex;
		}
		finally
		{
			// store the end date of the transfer
			endDate = System.currentTimeMillis();

			// the thread must be stopped, if it is running
			if ( thread != null )
			{
				thread.stop();
				thread = null;
			}

			// the transfer is terminated, the instant rate now have no sense
			transferRate = -1;

			processTransferTerminated( new DataTransferEvent( this ) );
		}
	}
	private void processDelayExpired( DataTransferEvent event )
	{
		if ( listeners != null )
		{
			for (int i = listeners.size() - 1; i >= 0; --i)
			{
				((DataTransferListener) listeners.get( i )).notificationDelayExpired( event );
			}
		}
	}
	private void processImpl()
		throws IOException
	{
		// allocate the buffer
		byte[] buffer = new byte[ notificationStep == -1 || notificationStep > 1024 ? 1024 : notificationStep ];

		// the amount of memory received since the previous notification
		int unotifiedDataLength = 0;

		// if the size of the data is known
		if ( expectedDataLength == -1 )
		{
			// will store the length of the buffer after a read process
			int len;

			while ( (len = read( inputStream, buffer, true, attemptCount )) != -1 )
			{
				write( outputStream, buffer, len, attemptCount );
				transferedDataLength += len;

				if ( notificationStep != -1 )
				{
					unotifiedDataLength += len;

					if ( unotifiedDataLength > notificationStep )
					{
						processStepProcessed( new DataTransferEvent( this ) );
						unotifiedDataLength -= notificationStep;
					}
				}

				// some transfers (ex : a local IDE hard drive to itself)
				// may use the whole CPU time, they must leave it to other tasks
				Thread.yield();
			}
		}
		else
		{
			// until the whole data has been transfered
			while ( transferedDataLength < expectedDataLength )
			{
				int len = read( inputStream, buffer, false, attemptCount );
				write( outputStream, buffer, len, attemptCount );
				transferedDataLength += len;

				if ( notificationStep != -1 )
				{
					unotifiedDataLength += len;

					if ( unotifiedDataLength > notificationStep )
					{
						processStepProcessed( new DataTransferEvent( this ) );
						unotifiedDataLength -= notificationStep;
					}
				}

				// some transfers (ex : a local IDE hard drive to itself)
				// may use the whole CPU time, they must leave it to other tasks
				Thread.yield();
			}
		}
	}
	private void processReadIOErrorOccured( DataTransferEvent event )
	{
		if ( listeners != null )
		{
			for (int i = listeners.size() - 1; i >= 0; --i)
			{
				((DataTransferListener) listeners.get( i )).readIOErrorOccured( event );
			}
		}
	}
	private void processStepProcessed( DataTransferEvent event )
	{
		if ( listeners != null )
		{
			for (int i = listeners.size() - 1; i >= 0; --i)
			{
				((DataTransferListener) listeners.get( i )).notificationStepProcessed( event );
			}
		}
	}
	private void processTransferStarted( DataTransferEvent event )
	{
		if ( listeners != null )
		{
			for (int i = listeners.size() - 1; i >= 0; --i)
			{
				((DataTransferListener) listeners.get( i )).transferStarted( event );
			}
		}
	}
	private void processTransferTerminated( DataTransferEvent event )
	{
		if ( listeners != null )
		{
			for (int i = listeners.size() - 1; i >= 0; --i)
			{
				((DataTransferListener) listeners.get( i )).transferTerminated( event );
			}
		}
	}
	private void processWriteIOErrorOccured( DataTransferEvent event )
	{
		if ( listeners != null )
		{
			for (int i = listeners.size() - 1; i >= 0; --i)
			{
				((DataTransferListener) listeners.get( i )).writeIOErrorOccured( event );
			}
		}
	}
	
	/**
	 * Reads data from the given input stream and store it into
	 * the given byte array. Listeners are notified of eventual I/O errors.
	 * @return the number of read bytes.
	 */
	private int read( InputStream inputStream, byte[] array, boolean allowEOF, int attemptCount )
		throws IOException
	{
		try
		{
			int len = inputStream.read( array );
			
			if ( len == -1 && !allowEOF )
			{
				throw new EOFException();
			}
			else
			{
				return len;
			}
		}
		catch ( EOFException ex )
		{
			throw ex;
		}
		catch ( IOException ex )
		{
			if ( attemptCount == 0 )
			{
				throw ex;
			}
			else
			{
				processReadIOErrorOccured( new DataTransferEvent( this ) );
				return read( inputStream, array, allowEOF, attemptCount - 1 );
			}
		}
	}
	public void removeDataTransferListener( DataTransferListener l )
	{
		listeners.remove( l );
		
		if ( listeners.isEmpty() )
		{
			listeners = null;
		}
	}
	public void run()
	{
		long delay = notificationDelay;
		int previousTransferedDataLength = transferedDataLength;

		for (;;)
		{
			try
			{
				// wait the appropriate delay
				Thread.sleep( delay );

				// evaluate the amount of data transfered during the wait period
				// this is the 'instantenous' transfer rate
				int len = transferedDataLength - previousTransferedDataLength;
				previousTransferedDataLength = transferedDataLength;

				// evaluate the transfer rate
				transferRate = (int) ((1000L * len) / delay);
				
				processDelayExpired( new DataTransferEvent( this ) );
			}
			catch ( InterruptedException ex )
			{
			}
		}
	}

	/**
	 * Sets the number of attempts to do when an error occur.
	 */
	public void setAttemptCount( int attemptCount )
	{
		if ( attemptCount < 0 )
			throw new IllegalArgumentException();

		this.attemptCount = attemptCount;
	}

	/**
	 * Sets the expected length of the datas (in bytes) to transfer or
	 * -1 if the length is unknown.
	 */
	public void setExpectedDataLength( int expectedDataLength )
	{
		if ( expectedDataLength < 0 )
			throw new IllegalArgumentException();

		this.expectedDataLength = expectedDataLength;
	}

	/**
	 * Sets the input stream the data is read from.
	 */
	public void setInputStream( InputStream inputStream )
	{
		if ( inputStream == null )
			throw new NullPointerException();

		this.inputStream = inputStream;
	}

	/**
	 * Sets the notification delay (in milliseconds).
	 * A good value is 1 seconde.
	 */
	public void setNotificationDelay( int notificationDelay )
	{
		if ( notificationDelay <= 0 )
			throw new IllegalArgumentException();

		this.notificationDelay = notificationDelay;
	}

	/**
	 * Sets the notification step (in bytes). An usual value is 1 kilo-byte
	 * for network transfers or 1 Mega-byte for local transfers.
	 */
	public void setNotificationStep( int notificationStep )
	{
		if ( notificationStep <= 0 )
			throw new IllegalArgumentException();

		this.notificationStep = notificationStep;
	}

	/**
	 * Sets the output stream the data is written to.
	 * If the output stream is null, the read data is lost (transfered nowhere).
	 */
	public void setOutputStream( OutputStream outputStream )
	{
		this.outputStream = outputStream;
	}

	/**
	 * Sets the total amount of transfered bytes.
	 */
	public void setTransferedDataLength( int transferedDataLength )
	{
		if ( transferedDataLength < 0 )
			throw new IllegalArgumentException();

		this.transferedDataLength = transferedDataLength;
	}

	/**
	 * Writes the given amount of data to the given output stream,
	 * getting it from the given byte array. Listeners are notified
	 * of eventual I/O errors.
	 */
	private void write( OutputStream outputStream, byte[] array, int length, int attemptCount )
		throws IOException
	{
		if ( outputStream != null )
		{
			try
			{
 				outputStream.write( array, 0, length );
			}
			catch ( IOException ex )
			{
				if ( attemptCount == 0 )
				{
					throw ex;
				}
				else
				{
					processWriteIOErrorOccured( new DataTransferEvent( this ) );
					write( outputStream, array, length, attemptCount - 1 );
				}
			}
		}
	}
}
