package toools.io.file;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import toools.OperatingSystem;
import toools.extern.Proces;

public abstract class AbstractFile
{
    File javaFile;
    private boolean temporary = false;

    public static String findUnusedNameIn(Directory location, String prefix, String suffix)
    {
	for (int i = 0;; ++i)
	{
	    String filename = prefix + i + suffix;
	    RegularFile f = new RegularFile(location, filename);

	    if (!f.exists())
	    {
		return filename;
	    }
	}
    }

    public AbstractFile(String path)
    {
	if (path.trim().isEmpty())
	    throw new IllegalArgumentException("empty file name");

	if (path.startsWith("~/"))
	{
	    path = Directory.getHomeDirectory().getPath() + '/' + path.substring(2);
	}
	else if (path.startsWith("$HOME/"))
	{
	    path = Directory.getHomeDirectory().getPath() + '/' + path.substring(6);
	}

	javaFile = new File(path).getAbsoluteFile();
    }

    private static String processPath(String p)
    {

	return p;
    }

    public boolean isTemporary()
    {
	return temporary;
    }

    public void setTemporary(boolean temporary)
    {
	this.temporary = temporary;
    }

    public boolean isNewerThan(AbstractFile f)
    {
	return getLastModificationDate() > f.getLastModificationDate();
    }

    public static <F extends AbstractFile> F map(File file, Class<F> expected)
    {
	if (file.exists())
	{
	    if (file.isDirectory())
	    {
		return (F) new Directory(file.getAbsolutePath());
	    }
	    else
	    {
		return (F) new RegularFile(file.getAbsolutePath());
	    }
	}
	else
	{
	    // we may be in the situation of a symbolic link targeting a
	    // non-existing file, so let's suppose this non-existing file was a regular
	    // files
	    if (expected == null)
	    {
		return (F) new RegularFile(file.getAbsolutePath());
		// throw new
		// IllegalArgumentException("file was expected to be found: " +
		// file.getAbsolutePath());
	    }
	    else
	    {
		try
		{
		    return expected.getConstructor(String.class).newInstance(file.getAbsolutePath());
		}
		catch (Throwable e)
		{
		    throw new IllegalStateException();
		}
	    }
	}
    }

    public static <F extends AbstractFile> F map(String path, Class<F> expected)
    {
	if (path.trim().isEmpty())
	    throw new IllegalArgumentException("empty file name");

	return map(new File(path).getAbsoluteFile(), expected);
    }

    @Override
    protected void finalize() throws Throwable
    {
	super.finalize();

	if (isTemporary())
	{
	    delete();
	}
    }

    /**
     * Computes the age of this file.
     * 
     * @return the age of this file, in milliseconds.
     */
    public long getAge()
    {
	return System.currentTimeMillis() - getLastModificationDate();
    }

    /*
     * public boolean isAbsolute() { return jf.isAbsolute(); }
     */
    public long getLastModificationDate()
    {
	if (!exists())
	    throw new IllegalStateException("cannot get the date of a non-existing file " + this);

	return javaFile.lastModified();
    }

    public void setLastModified(long i)
    {
	this.javaFile.setLastModified(i);

    }

    public String getName()
    {
	return javaFile.getName();
    }

    public String getPathRelativeToCurrentDir()
    {
	return getNameRelativeTo(Directory.getCurrentDirectory());
    }

    public String getNameRelativeTo(Directory d)
    {
	// if this file is somewhere in a subdirection of d
	if (getPath().startsWith(d.getPath()))
	{
	    // simply remove the path to d
	    return getPath().substring(d.getPath().length() + File.separator.length());
	}
	else
	{
	    return getPath();
	}
    }

 

    public abstract long getSize();

    public abstract void delete();

    public Directory getParent()
    {
	return new Directory(javaFile.getParent());
    }

    @Override
    public String toString()
    {
	Map<String, Object> l = new HashMap<String, Object>();
	l.put("name", getPath());
	l.put("type", this instanceof Directory ? "directory" : "plain file");
	l.put("exist", exists());

	if (exists())
	{
	    l.put("size", getSize());
	}

	return l.toString();
    }

    public String getPath()
    {
	return javaFile.getAbsolutePath();
    }

    public boolean exists()
    {
	return javaFile.exists();
    }

    @Override
    public boolean equals(Object obj)
    {
	return obj instanceof AbstractFile && equals((AbstractFile) obj);
    }

    public boolean equals(AbstractFile f)
    {
	return getPath().equals(f.getPath());
    }

    @Override
    public int hashCode()
    {
	return getPath().hashCode();
    }

    public void renameTo(String newPath) throws IOException
    {
	File newFile = new File(newPath).getAbsoluteFile();
	boolean ok = javaFile.renameTo(newFile);
	
	// if renaming failed, try copying+deleting
	if (!ok)
	{
	    throw new IOException("failed at renaming file " + getPath() + " to " + newFile.getAbsolutePath());
	}

	javaFile = newFile;
    }

    public static void main(String[] args) throws IOException
    {
	RegularFile f = new RegularFile("/tmp/vfkdsjh");
	f.setContent("coucou".getBytes());
	System.out.println(f);
	f.renameTo("/sdfsd");
	System.out.println(f);
    }

    public File toFile()
    {
	return this.javaFile;
    }

    public boolean canRead()
    {
	if (!exists())
	    throw new IllegalStateException("cannot get the status of a non-existing file " + this.getPath());

	return javaFile.canRead();
    }

    public boolean canWrite()
    {
	if (!exists())
	{
	    return getParent().canWrite();
	}
	else
	{
	    return javaFile.canWrite();
	}
    }

    public boolean isExecutable()
    {
	if (!exists())
	    throw new IllegalStateException("cannot get the status of a non-existing file " + this.getPath());

	return javaFile.canExecute();
    }

    public void setExecutable(boolean b)
    {
	javaFile.setExecutable(b);
    }

    public URL toURL()
    {
	try
	{
	    return this.javaFile.toURI().toURL();
	}
	catch (MalformedURLException e)
	{
	    e.printStackTrace();
	    throw new IllegalStateException();
	}
    }

    public boolean isEmpty()
    {
	return getSize() == 0;
    }

    public abstract boolean create() throws IOException;

    /**
     * Opens the file using the OS specific strategy.
     */
    public void open()
    {
	if (OperatingSystem.isMacOSX())
	{
	    Proces.exec("open", getPath());
	}
	else if (OperatingSystem.isLinux())
	{
	    Proces.exec("firefox", getPath());
	}
	else if (OperatingSystem.isWindows())
	{
	    Proces.exec("firefox", getPath());
	}
    }
}
