package toools.io.file;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;


import toools.OperatingSystem;
import toools.extern.Proces;
import toools.io.ScannerListener;
import toools.io.Utilities;
import toools.text.TextUtilities;

public class Directory extends AbstractFile
{
    private static final Directory tempDirectory = new Directory(System.getProperty("java.io.tmpdir"));

    static
    {
	if (!tempDirectory.exists())
	{
	    tempDirectory.mkdirs();
	}
    }

    public static Directory getSystemTempDirectory()
    {
	return tempDirectory;
    }

    public static Directory createTempDirectory(Directory location, String prefix, String suffix)
    {
	return new Directory(location, AbstractFile.findUnusedNameIn(location, prefix, suffix));
    }

    public Directory(String path)
    {
	super(path);

	if (javaFile.exists() && !javaFile.isDirectory())
	    throw new IllegalArgumentException(path + " is not a directory " + this.getPath());
    }

    public Directory(Directory parent, String name)
    {
	this(parent.getPath() + '/' + name);
    }

    @Override
    public void delete()
    {
	if (!exists())
	    throw new IllegalStateException("cannot delete a non-existing directory " + this.getPath());

	javaFile.delete();

	if (exists())
	    throw new IllegalStateException("directory should no longer exist " + this.getPath());
    }

    public void deleteRecursively()
    {
	if (!exists())
	    throw new IllegalStateException("cannot delete a non-existing directory " + this.getPath());

	for (AbstractFile thisChild : getChildFiles())
	{
	    if (thisChild instanceof Directory)
	    {
		((Directory) thisChild).deleteRecursively();
	    }
	    else
	    {
		thisChild.delete();
	    }
	}

	javaFile.delete();

	if (exists())
	    throw new IllegalStateException("directory should no longer exist " + this.getPath());
    }

    public void copyTo(Directory destination, boolean overwrite) throws FileNotFoundException, IOException
    {
	if (!exists())
	    throw new IllegalStateException("cannot copy a non-existing directory " + this.getPath());

	if (!overwrite && destination.exists())
	    throw new IllegalStateException("don't want to overwrite " + destination);

	destination.mkdirs();

	for (AbstractFile c : getChildFiles())
	{
	    if (c instanceof RegularFile)
	    {
		((RegularFile) c).copyTo(new RegularFile(destination.getPath() + '/' + c.getName()), overwrite);
	    }
	    else if (c instanceof Directory)
	    {
		((Directory) c).copyTo(new Directory(destination.getPath() + '/' + c.getName()), overwrite);
	    }
	    else
	    {
		throw new IllegalStateException("don't know this king of file " + c.getClass());
	    }
	}
    }

    @Override
    public long getSize()
    {
	if (!exists())
	    throw new IllegalStateException("cannot get the size of a non-existing directory " + this.getPath());

	long sum = 0;

	for (AbstractFile thisChild : getChildFiles())
	{
	    if (thisChild.exists())
	    {
		    sum += thisChild.getSize();
	    }
	}

	return sum;
    }

    public Set<AbstractFile> getChildFiles()
    {
	if (!exists())
	    throw new IllegalStateException("cannot get the children of a non-existing directory " + this.getPath());

	Set<AbstractFile> children = new HashSet<AbstractFile>();

	for (String s : javaFile.list())
	{
	    children.add(AbstractFile.map(getPath() + '/' + s, null));
	}

	return children;
    }

    public boolean mkdirs()
    {
	if (exists())
	    throw new IllegalStateException("cannot create and already-existing directory " + this.getPath());

	return javaFile.mkdirs();
    }

    public List<AbstractFile> retrieveTree()
    {
	return retrieveTree(null, null);
    }

    public List<AbstractFile> retrieveTree(FileFilter filter, ScannerListener l)
    {
	List<AbstractFile> files = new ArrayList<AbstractFile>();

	if (Utilities.operatingSystemIsUNIX())
	{
	    for (String line : TextUtilities.splitInLines(new String(Proces.exec("find", getPath()))))
	    {
		AbstractFile f = AbstractFile.map(line, null);

		if (filter == null || filter.accept(f))
		{
		    files.add(f);

		    if (l != null)
		    {
			l.foundFile(f);
		    }
		}
	    }
	}
	else
	{
	    Stack<Directory> stack = new Stack<Directory>();
	    stack.push(this);

	    while (!stack.isEmpty())
	    {
		Directory f = stack.pop();

		for (AbstractFile child : f.getChildFiles())
		{
		    if (filter == null || filter.accept(child))
		    {
			files.add(child);

			if (l != null)
			{
			    l.foundFile(child);
			}
		    }

		    if (child instanceof Directory)
		    {
			stack.push((Directory) child);
		    }
		}
	    }
	}

	return files;
    }

    public List<AbstractFile> getChildFiles(FileFilter... ffs)
    {
	List<AbstractFile> l = new ArrayList<AbstractFile>();

	for (AbstractFile c : getChildFiles())
	{
	    if (accept(c, ffs))
	    {
		l.add(c);
	    }
	}

	return l;
    }

    private boolean accept(AbstractFile f, FileFilter... ffs)
    {
	for (FileFilter ff : ffs)
	{
	    if (!ff.accept(f))
	    {
		return false;
	    }
	}

	return true;
    }

    public List<Directory> listDirectories()
    {
	return (List<Directory>) (List) getChildFiles(new FileFilter.DirectoryFilter());
    }

    public List<RegularFile> listRegularFiles()
    {
	return (List<RegularFile>) (List) getChildFiles(new FileFilter.RegularFileFilter());
    }

    public <F extends AbstractFile> F getChild(String string, Class<F> defaultClass)
    {
	return AbstractFile.map(getPath() + '/' + string, defaultClass);
    }

    public List<AbstractFile> findChildFilesWhoseTheNameMatches(String re)
    {
	return getChildFiles(new FileFilter.RegexFilter(re));
    }

    public String getUniqFileName(String prefix, String suffix)
    {
	for (int i = 0;; ++i)
	{
	    String name = prefix + i + suffix;

	    if (!new File(getPath(), name).exists())
	    {
		return name;
	    }
	}
    }

    public Directory getChildDirectory(String name)
    {
	return getChild(name, Directory.class);
    }

    public RegularFile getChildRegularFile(String name)
    {
	return getChild(name, RegularFile.class);
    }

    public List<RegularFile> getChildRegularFilesMatching(String re)
    {
	return (List<RegularFile>) (List) getChildFiles(new FileFilter.RegularFileFilter(), new FileFilter.RegexFilter(
		re));
    }

    public static Directory getCurrentDirectory()
    {
	return new Directory(System.getProperty("user.dir"));
    }

    public static Directory getHomeDirectory()
    {
	return new Directory(System.getProperty("user.home"));
    }

    @Override
    public boolean create() throws IOException
    {
	return mkdirs();
    }

}
