package toools.io.file;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import toools.io.Utilities;
import toools.text.TextUtilities;

public class RegularFile extends AbstractFile
{

    public RegularFile(String path)
    {
	super(path);
    }

    public RegularFile(Directory parent, String name)
    {
	this(parent.getPath() + '/' + name);
    }

    public static RegularFile createTempFile()
    {
	return createTempFile("toools-", ".tmp");
    }

    public static RegularFile createTempFile(String prefix, String suffix)
    {
	try
	{
	    return createTempFile(Directory.getSystemTempDirectory(), prefix, suffix);
	}
	catch (IOException e)
	{
	    throw new IllegalStateException("cannot create file in temp directory: "
		    + Directory.getSystemTempDirectory());
	}
    }

    public static RegularFile createTempFile(Directory location, String prefix, String suffix) throws IOException
    {
	RegularFile f = new RegularFile(location, AbstractFile.findUnusedNameIn(location, prefix, suffix));
	f.create();
	return f;
    }

    public void setContent(byte[] bytes) throws IOException
    {
	BufferedOutputStream bos = createWritingStream();
	bos.write(bytes);
	bos.close();
    }

    public void setContent(String text, Charset charset) throws IOException
    {
	setContent(text.getBytes(charset));
    }

    public void setContentAsUTF8(String text) throws IOException
    {
	setContent(text.getBytes("UTF-8"));
    }

    public void setContentAsUTF16(String text) throws IOException
    {
	setContent(text.getBytes("UTF-16"));
    }

    public void setContentAsASCII(String text) throws IOException
    {
	setContent(text.getBytes("US-ASCII"));
    }

    public void setContentAsMacRoman(String text) throws IOException
    {
	setContent(text.getBytes("MacRoman"));
    }

    public byte[] getContent() throws IOException
    {
	BufferedInputStream bis = createReadingStream();
	byte[] bytes = Utilities.readUntilEOF(bis);
	bis.close();
	return bytes;
    }

    public void copyTo(RegularFile destination, boolean overwrite) throws FileNotFoundException, IOException
    {
	if (!exists())
	    throw new IllegalStateException("cannot copy a non-existing file " + this.getPath());

	if (!overwrite && destination.exists())
	    throw new IllegalStateException("don't want to overwrite " + destination.getPath());

	InputStream is = createReadingStream();
	OutputStream os = destination.createWritingStream();
	Utilities.copy(is, os);
	is.close();
	os.close();
    }

    public void moveTo(RegularFile destination, boolean overwrite) throws FileNotFoundException, IOException
    {
	if (!exists())
	    throw new IllegalStateException("cannot move a non-existing file " + this.getPath());

	copyTo(destination, overwrite);
	delete();
    }

    @Override
    public void delete()
    {
	if (!exists())
	    throw new IllegalStateException("cannot delete a non-existing file " + this.getPath());

	javaFile.delete();
    }

    @Override
    public long getSize()
    {
	if (!exists())
	    throw new IllegalStateException("cannot get the size of a non-existing file " + this.getPath());

	return javaFile.length();
    }

    public BufferedOutputStream createWritingStream() throws FileNotFoundException
    {
	return createWritingStream(false);
    }

    public BufferedOutputStream createWritingStream(boolean append) throws FileNotFoundException
    {
	if (!getParent().exists())
	    throw new FileNotFoundException("parent directory does not exist: " + getParent().getPath());

	return new BufferedOutputStream(new FileOutputStream(javaFile, append));
    }

    public BufferedInputStream createReadingStream() throws FileNotFoundException
    {
	if (!exists())
	    throw new FileNotFoundException("cannot read a non-existing file " + this.getPath());

	return new BufferedInputStream(new FileInputStream(javaFile));
    }

    public static boolean sameContents(RegularFile a, RegularFile b) throws IOException
    {
	if (a.getSize() != b.getSize())
	{
	    return false;
	}
	else
	{
	    return compareFileContentsLexicographically(a, b) == 0;
	}
    }

    public static int compareFileContentsLexicographically(RegularFile a, RegularFile b) throws IOException
    {
	if (a.getSize() == 0 && b.getSize() == 0)
	{
	    return 0;
	}
	else if (a.getSize() == 0)
	{
	    return -1;
	}
	else if (b.getSize() == 0)
	{
	    return 1;
	}
	else
	{
	    return TextUtilities.compareLexicographically(a.createReadingStream(), b.createReadingStream());
	}
    }

    public String getNameWithoutExtension()
    {
	return getName().substring(0, getName().length() - getExtension().length() - 1);
    }

    @Override
    public boolean create() throws IOException
    {
	return javaFile.createNewFile();
    }

    public List<String> getLines() throws IOException
    {
	return TextUtilities.splitInLines(new String(getContent()));
    }

    public boolean isArchive()
    {
	Collection<String> extensions = Arrays.asList("tgz", "tar.gz", "zip");
	return extensions.contains(getExtension());
    }

    public String getExtension()
    {
	List<String> extensions = getExtensions();

	if (extensions.isEmpty())
	{
	    return null;
	}
	else
	{
	    return extensions.get(extensions.size() - 1);
	}
    }

    public List<String> getExtensions()
    {
	List<String> extensions = new ArrayList();
	String name = getName();

	while (true)
	{
	    int pos = name.lastIndexOf('.');

	    if (pos > 0)
	    {
		extensions.add(0, name.substring(pos + 1));
		name = name.substring(0, pos);
	    }
	    else
	    {
		return extensions;
	    }
	}
    }

    public void setExtension(String newExtension) throws IOException
    {
	String ext = getExtension();
	String newName = getPath().replaceFirst(ext + "$", newExtension);
	renameTo(newName);
    }

    public void append(byte[] bytes) throws IOException
    {
	BufferedOutputStream os = createWritingStream(true);
	os.write(bytes);
	os.close();
    }

}
