package toools.math;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;



import toools.extern.Proces;
import toools.gui.PDFRenderingAWTComponent;
import toools.io.FileUtilities;
import toools.io.file.Directory;
import toools.io.file.RegularFile;

public class Distribution<T>
{
    private final Map<T, Long> map = new HashMap<T, Long>();
    private long sum = 0;
    private long greatestNumberOfOccurences = 0;
    private final String title;

    public Distribution(String title)
    {
	this.title = title;
    }

    public long getGreatestNumberOfOccurences()
    {
	return greatestNumberOfOccurences;
    }

    public void addOccurence(T t)
    {
	addNOccurences(t, 1);
    }

    public void addNOccurences(T e, long add)
    {
	if (add < 0)
	    throw new IllegalArgumentException();

	Long n = map.get(e);

	if (n == null)
	{
	    map.put(e, add);
	    sum += add;

	    if (add > this.greatestNumberOfOccurences)
	    {
		greatestNumberOfOccurences = add;
	    }
	}
	else
	{
	    long newValue = n + add;
	    map.put(e, newValue);
	    sum += add;

	    if (newValue > this.greatestNumberOfOccurences)
	    {
		greatestNumberOfOccurences = newValue;
	    }
	}
    }

    public long getNumberOfOccurences(T t)
    {
	if (!map.containsKey(t))
	    throw new IllegalArgumentException("element is not considered");

	return map.get(t);
    }

    public long getTotalNumberOfOccurences()
    {
	return sum;
    }

    public double getRelativeNumberOfOccurences(T t)
    {
	return (double) getNumberOfOccurences(t) / (double) getTotalNumberOfOccurences();
    }

    public List<T> orderObjectsByIncreasingNumberOfOccurences()
    {
	List<T> l = new ArrayList<T>(getOccuringObjects());

	Collections.sort(l, new Comparator<T>() {

	    public int compare(T o1, T o2)
	    {
		return new Long(getNumberOfOccurences(o1)).compareTo(getNumberOfOccurences(o2));
	    }
	});

	return l;
    }

    public Collection<T> getOccuringObjects()
    {
	return Collections.unmodifiableCollection(this.map.keySet());
    }

    public String toGNUPlotData(List<T> list, boolean relative)
    {
	StringBuilder b = new StringBuilder();

	for (T o : list)
	{
	    b.append(o.toString() + '\t' + (relative ? getRelativeNumberOfOccurences(o) : getNumberOfOccurences(o))
		    + '\n');
	}

	return b.toString();
    }

    public String toGNUPlotData(boolean relative)
    {
	return toGNUPlotData(new ArrayList(getOccuringObjects()), relative);
    }

    public String getGNUPlotCommands(String datafileName, boolean relative)
    {
	StringBuilder b = new StringBuilder();
	b.append("set term pdf\nset style data histogram\nset style fill solid border -1\nset style data histogram\n");
	b.append("set boxwidth\nset auto x\n");
	// b.append("set boxwidth 1\n");
	b.append("set yrange [0:" + (relative ? 1 : (1.1 * getGreatestNumberOfOccurences())) + "]\n");
	b.append("plot \"" + datafileName + "\" with  impulses title \"" + title + "\"\n");
	return b.toString();
    }

    public byte[] toPDF()
    {
	try
	{
	    RegularFile datafile = FileUtilities.createTempFile("toools-", ".dat");
	    datafile.setContent(toGNUPlotData(true).getBytes());
	    // System.out.println(getGNUPlotCommands("hehe", true));
	    byte[] pdf = Proces.exec("gnuplot", getGNUPlotCommands(datafile.getPath(), true).getBytes());
	    datafile.delete();
	    return pdf;
	}
	catch (IOException e)
	{
	    e.printStackTrace();
	    throw new IllegalStateException();
	}
    }

    public static Distribution<Integer> getDistribution(String title, int[] values)
    {
	Distribution<Integer> distribution = new Distribution<Integer>(title);

	for (int v : values)
	{
	    distribution.addOccurence(v);
	}

	return distribution;
    }

    public static Distribution<Long> getDistribution(String title, long[] values)
    {
	Distribution<Long> distribution = new Distribution<Long>(title);

	for (long v : values)
	{
	    distribution.addOccurence(v);
	}

	return distribution;
    }

    public static void main(String[] args) throws Exception
    {
	Distribution<Integer> d = new Distribution<Integer>("RANDOM example");
	Random r = new Random();
	for (int i = 0; i < 20; ++i)
	{
	    d.addOccurence(r.nextInt(10));
	}

	d.display();

    }

    public String getTitle()
    {
	return title;
    }

    public void display()
    {
	PDFRenderingAWTComponent c = new PDFRenderingAWTComponent();
	toools.gui.Utilities.displayInJFrame(c, "Distribution");
	c.setPDFData(toPDF(), 0);
    }

    @Override
    public String toString()
    {
	StringBuilder b = new StringBuilder();

	for (T t : getOccuringObjects())
	{
	    b.append(t);
	    b.append('\t');
	    b.append(getNumberOfOccurences(t));
	    b.append('\n');
	}

	return b.toString();
    }

}
