package toools.math;

import java.io.File;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import toools.NotYetImplementedException;
import toools.collections.Collections;
import toools.io.FileUtilities;
import toools.text.TextUtilities;


/**
 * @author luc.hogie Created on Jun 14, 2004
 */
public class MathsUtilities
{

    public static <T> Collection<Couple<T>> computeCouples(Collection<T> c1, Collection<T> c2)
    {
	Collection<Couple<T>> couples = new HashSet<Couple<T>>();

	for (T t1 : c1)
	{
	    for (T t2 : c2)
	    {
		if (t1 != t2)
		{
		    couples.add(new Couple<T>(t1, t2));
		}
	    }
	}

	return couples;
    }

	/**
	 * Computes the number of subsets of k elements in a set of n elements (the binomial).
	 * @param n set of n elements
	 * @param k size of the subsets
	 * @return number of subsets
	 */
	public static long binomial(int n, int k)
	{
		long[] b = new long[n + 1];
		b[0] = 1;

		for (int i = 1; i <= n; ++i)
		{
			b[i] = 1;

			for (int j = i - 1; j > 0; --j)
			{
				b[j] += b[j - 1];
			}
		}
		
		return b[k];
	}
	
	/**
	 * Computes the standard deviation of a collection of long values.
	 * @param c the collection of long values
	 * @return the standard deviation
	 */
	public static double computeLongStandardDeviation(Collection<Long> c)
	{
		double avg = MathsUtilities.computeLongAverage(c);
		double diffSum = 0;

		for (long n : c)
		{
			diffSum += Math.abs(avg - n);
		}

		return diffSum / c.size();
	}

    
    public static long computeLongAverage(long[] values)
    {
	if (values.length == 0)
	    throw new IllegalArgumentException("tab is empty");

	return getSum(values) / values.length;
    }

    public static long getSum(long[] values)
    {
	if (values.length == 0)
	    throw new IllegalArgumentException("tab is empty");

	long sum = 0;

	for (long n : values)
	{
	    sum += n;
	}

	return sum;
    }

    public static int computeMinimum(int[] a)
    {
	int min = a[0];

	for (int i = 1; i < a.length; ++i)
	{
	    int b = a[i];

	    if (b < min)
	    {
		min = b;
	    }

	}

	return min;
    }

    public static long computeMinimum(long[] a)
    {
	long min = a[0];

	for (int i = 1; i < a.length; ++i)
	{
	    long b = a[i];

	    if (b < min)
	    {
		min = b;
	    }

	}

	return min;
    }

    public static int computeMaximum(int[] a)
    {
	int max = a[0];

	for (int i = 1; i < a.length; ++i)
	{
	    int b = a[i];

	    if (b > max)
	    {
		max = b;
	    }

	}

	return max;
    }

    public static double computeMaximum(double[] a)
    {
	double max = a[0];

	for (int i = 1; i < a.length; ++i)
	{
	    double b = a[i];

	    if (b > max)
	    {
		max = b;
	    }

	}

	return max;
    }

    public static String toGNUPlotText(List<? extends Number> list)
    {
	StringBuffer buf = new StringBuffer();

	for (Number n : list)
	{
	    buf.append(n.toString());
	    buf.append('\n');
	}

	return buf.toString();
    }

    public static byte[] toPDF(File inputFile, File outputFile)
    {
	String gnuplotText = "set term postscript" + "set output \"${OUTPUT}\"" + "set xlabel \"${XLABEL}"
		+ "set ylabel \"${YLABEL}\"" + "set key left top" + "plot \"${INPUT}\" with linespoints";

	gnuplotText.replace("${INPUT}", inputFile.getAbsolutePath());
	gnuplotText.replace("${OUTPUT}", outputFile.getAbsolutePath());

	File gnuplotFile = FileUtilities.getUniqFile("kj", "kj.gnuplot");
	// FileUtilities.setFileContent(gnuplotFile, gnuplotText.getBytes());
	return null;

    }

    public static double getCumulatedVariations(List<Double> values)
    {
	double variation = 0;

	for (int i = 1; i < values.size(); ++i)
	{
	    variation += Math.abs(values.get(i) - values.get(i - 1));
	}

	return variation;
    }

    public static double nPercentOf(double percentage, double amount)
    {
	return (amount * percentage) / 100d;
    }

    public static double round(double n, int precision)
    {
	double a = Math.pow(10, precision);
	n *= a;
	n = Math.round(n);
	n /= a;
	return n;
    }

    public static double pickRandomBetween(double min, double max, Random random)
    {
	if (min == max)
	{
	    return min;
	}
	else if (min < max)
	{
	    return random.nextDouble() * (max - min) + min;
	}
	else
	{
	    throw new IllegalArgumentException("min=" + min + ", max=" + max);
	}
    }

    /**
     * in inclusive out exclusive
     * 
     * @param min
     * @param max
     * @param random
     * @return
     */
    public static int pickRandomBetween(int min, int max, Random random)
    {
	if (min >= max)
	    throw new IllegalArgumentException("min=" + min + ", max=" + max);

	return random.nextInt(max - min) + min;
    }

    public static double cut(double a, int precision)
    {
	double m = Math.pow(10, precision);
	a *= m;
	a = (int) a;
	a /= m;
	return a;
    }

    /*
     * Return the angle of the given vector
     */
    public static double getAngleForVector(double i, double j)
    {
	return Math.atan(j / i);
    }

    public static void main(String[] args)
    {
	int[] ind = new int[3];
	int[] s = new int[] {2, 5, 7};
	
	while (next(ind, s))
	{
	    for (int i : ind)
	    {
		    System.out.print(i + ", ");
	    }
	    
	    System.out.println();
	}
    }

    public static double getMinimum(Collection<Double> c)
    {
	if (c.isEmpty())
	    throw new IllegalArgumentException("set is empty");

	Iterator<Double> i = c.iterator();
	double min = i.next();

	while (i.hasNext())
	{
	    double d = i.next();

	    if (d < min)
	    {
		min = d;
	    }
	}

	return min;
    }

    public static double getMaximum(Collection<Double> c)
    {
	if (c.isEmpty())
	    throw new IllegalArgumentException("set is empty");

	Iterator<Double> i = c.iterator();
	double max = i.next();

	while (i.hasNext())
	{
	    double d = i.next().doubleValue();

	    if (d > max)
	    {
		max = d;
	    }
	}

	return max;
    }

    public static int getMaximum(int[] array)
    {
	if (array.length == 0)
	    throw new IllegalArgumentException("set is empty");

	int max = array[0];

	for (int i = 1; i < array.length; ++i)
	{
	    if (array[i] > max)
	    {
		max = array[i];
	    }
	}

	return max;
    }

    public static int getMinimum(int[] array)
    {
	if (array.length == 0)
	    throw new IllegalArgumentException("set is empty");

	int min = array[0];

	for (int i = 1; i < array.length; ++i)
	{
	    if (array[i] < min)
	    {
		min = array[i];
	    }
	}

	return min;
    }

    public static double getSum(Collection<Double> c)
    {
	if (c.isEmpty())
	    throw new IllegalArgumentException("set is empty");

	double sum = 0;

	for (double n : c)
	{
	    sum += n;
	}

	return sum;
    }

    public static double getVariationSum(Collection<? extends Number> c)
    {
	if (c.isEmpty())
	    throw new IllegalArgumentException("set is empty");

	double sum = 0;
	Iterator<? extends Number> i = c.iterator();
	double previous = i.next().doubleValue();

	while (i.hasNext())
	{
	    double v = i.next().doubleValue();
	    sum += Math.abs(v - previous);
	}

	return sum;
    }

    public static double computeAverage(Collection<Double> c)
    {
	return getSum(c) / c.size();
    }

    public static double computeLongAverage(Collection<Long> values)
    {
	double avg = 0;
	// add the mean and std value to the mean history
	for (long a : values)
	{
	    avg += a;
	}

	return avg / values.size();
    }

    public static double computeIntegerAverage(Collection<Integer> values)
    {
	double avg = 0;
	// add the mean and std value to the mean history
	for (int a : values)
	{
	    avg += a;
	}

	return avg / values.size();
    }

    public static long computeLongStd(Collection<Long> values, double avg)
    {
	long stdDeviation = 0;

	for (long n : values)
	{
	    stdDeviation += Math.abs(avg - n);
	}

	stdDeviation += stdDeviation / values.size();

	return stdDeviation;
    }

    public static double computeDoubleStd(Collection<Double> values, double avg)
    {
	double stdDeviation = 0;

	for (double n : values)
	{
	    stdDeviation += Math.abs(avg - n);
	}

	stdDeviation += stdDeviation / values.size();

	return stdDeviation;
    }

    public static int computeIntegerStd(Collection<Integer> values, double avg)
    {
	int stdDeviation = 0;

	for (int n : values)
	{
	    stdDeviation += Math.abs(avg - n);
	}

	stdDeviation += stdDeviation / values.size();

	return stdDeviation;
    }

    public static double computeAverage(int[] array)
    {
	return (double) getSum(array) / (double) array.length;
    }

    public static double computeAverage(double[] array)
    {
	return (double) getSum(array) / (double) array.length;
    }

    public static double computeSum(DoubleIterator p)
    {
	double a = 0;

	while (p.hasNext())
	{
	    a += p.next();
	}

	return a;
    }

    public static double computeMinimum(double[] a)
    {
	double min = a[0];

	for (int i = 1; i < a.length; ++i)
	{
	    double b = a[i];

	    if (b < min)
	    {
		min = b;
	    }

	}

	return min;
    }

    public static double computeMinimum(DoubleIterator p)
    {
	if (!p.hasNext())
	    throw new IllegalStateException();

	double a = p.next();

	while (p.hasNext())
	{
	    double b = p.next();

	    if (b < a)
	    {
		a = b;
	    }
	}

	return a;
    }

    public static int computeMinimum(IntIterator p)
    {
	if (!p.hasNext())
	    throw new IllegalArgumentException("no value to compute");

	int min = p.next();

	while (p.hasNext())
	{
	    int b = p.next();

	    if (b < min)
	    {
		min = b;
	    }
	}

	return min;
    }

    public static int computeMaximum(IntIterator p)
    {
	if (!p.hasNext())
	    throw new IllegalArgumentException("no value to compute");

	int max = p.next();

	while (p.hasNext())
	{
	    int b = p.next();

	    if (b > max)
	    {
		max = b;
	    }
	}

	return max;
    }

    public static double computeAverage(DoubleIterator p)
    {
	double a = 0;

	while (p.hasNext())
	{
	    double n = p.next();
	}

	throw new IllegalAccessError();
    }

    public static double getSum(double[] array)
    {
	if (array.length == 0)
	    throw new IllegalArgumentException("nothing to sum");

	double sum = 0;

	for (int i = 0; i < array.length; ++i)
	{
	    sum += array[i];
	}

	return sum;
    }

    public static int getSum(int[] array)
    {
	if (array.length == 0)
	    throw new IllegalArgumentException("nothing to sum");

	int sum = 0;

	for (int i = 0; i < array.length; ++i)
	{
	    sum += array[i];
	}

	return sum;
    }

    public static double computeStandardDeviation(Collection<Double> c)
    {
	double avg = computeAverage(c);
	double diffSum = 0;

	for (double n : c)
	{
	    diffSum += Math.abs(avg - n);
	}

	return diffSum / c.size();
    }

    public static double computeStandardDeviation(double[] c)
    {
	double avg = computeAverage(c);
	double diffSum = 0;

	for (double n : c)
	{
	    diffSum += Math.abs(avg - n);
	}

	return diffSum / c.length;
    }

    public static List<Double> getAverageList(Collection<List<Double>> lists)
    {
	List<Double> resList = new Vector<Double>();

	if (lists.size() == 0)
	{
	    throw new IllegalArgumentException("no lists");
	}
	else
	{
	    int size = lists.iterator().next().size();

	    for (int i = 0; i < size; ++i)
	    {
		Collection<Double> values = Collections.getElementsAt(lists, i);
		double avg = computeAverage(values);
		resList.add(new Double(avg));
	    }
	}

	return resList;
    }

    public static List<Double> getStandardDeviationList(Collection<List<Double>> lists)
    {
	List<Double> resList = new Vector<Double>();

	if (lists.size() == 0)
	{
	    throw new IllegalArgumentException("no lists");
	}
	else
	{
	    int size = lists.iterator().next().size();

	    for (int i = 0; i < size; ++i)
	    {
		Collection<Double> values = Collections.getElementsAt(lists, i);
		double avg = computeStandardDeviation(values);
		resList.add(new Double(avg));
	    }
	}

	return resList;
    }

    private static boolean collectionsHaveSameSize(Collection<Collection> lists)
    {
	if (lists.isEmpty())
	{
	    throw new IllegalArgumentException("no lists");
	}
	else
	{
	    Iterator<Collection> i = lists.iterator();
	    int ref = i.next().size();

	    while (i.hasNext())
	    {
		int sz = i.next().size();

		if (sz != ref)
		{
		    return false;
		}
	    }

	    return true;
	}
    }

    public static boolean isNumber(String s)
    {
	return TextUtilities.isDouble(s);
    }

    public static double getClosestValue(double ref, Collection<Double> c)
    {
	if (c.isEmpty())
	{
	    throw new IllegalArgumentException("no values to choose from");
	}
	else if (c.size() == 1)
	{
	    return c.iterator().next();
	}
	else
	{
	    Iterator<Double> i = c.iterator();
	    double closest = i.next();

	    while (i.hasNext())
	    {
		double d = i.next();

		// if the difference is smaller
		if (Math.abs(ref - d) < Math.abs(ref - closest))
		{
		    closest = d;
		}
	    }

	    return closest;
	}
    }

    public static boolean isInteger(double value)
    {
	int intValue = (int) value;
	return value == (double) intValue;
    }

    public static int compare(long a, long b)
    {
	if (a == b)
	{
	    return 0;
	}
	else if (a < b)
	{
	    return -1;
	}
	else
	{
	    return 1;
	}
    }

    public static int compare(int a, int b)
    {
	if (a == b)
	{
	    return 0;
	}
	else if (a < b)
	{
	    return -1;
	}
	else
	{
	    return 1;
	}
    }

    public static int compare(double a, double b)
    {
	if (a == b)
	{
	    return 0;
	}
	else if (a < b)
	{
	    return -1;
	}
	else
	{
	    return 1;
	}
    }

    /**
     * Compare the two given object, if possible.
     * 
     * @param e1
     * @param e2
     * @return
     */
    public static int compare(Object e1, Object e2)
    {
	Class<?> clazz = e1.getClass();

	if (clazz == e2.getClass())
	{

	    if (clazz == Integer.class)
	    {
		return ((Integer) e1).compareTo((Integer) e2);
	    }
	    else if (clazz == Double.class)
	    {
		return ((Double) e1).compareTo((Double) e2);
	    }
	    else if (clazz == String.class)
	    {
		return ((String) e1).compareTo((String) e2);
	    }
	    else
	    {
		throw new IllegalArgumentException("elements are not comparable: " + e1.getClass());
	    }
	}
	else
	{
	    throw new IllegalArgumentException("not the same class: " + e1.getClass() + " != " + e2.getClass());
	}
    }

    public static int max(int[] a)
    {
	int max = -1;

	for (int i = 0; i < a.length; ++i)
	{
	    int d = a[i];

	    if (d > max)
	    {
		max = d;
	    }
	}
	return max;
    }

    public static void shuffle(int[] a)
    {
	int[] indexes = new int[a.length];

	for (int i = 0; i < indexes.length; ++i)
	{
	}
	throw new NotYetImplementedException();
    }
   
    public static boolean next(int[] values, int[] sizes)
    {
	if (values.length != sizes.length)
		throw new IllegalArgumentException("not same array size");

	if (values.length == 0)
		throw new IllegalArgumentException("empty arrays are not allowed");

	for (int s : sizes)
	    if (s <= 0)
		throw new IllegalArgumentException("size must be > 0");

	for (int i : values)
	    if (i < 0)
		throw new IllegalArgumentException("negative indices are not allowed");

	int i = 0;

	// while the value at this index is max
	while (values[i] == sizes[i] - 1)
	{
	    if (i == values.length - 1)
	    {
		return false;
	    }
	    else
	    {
		values[i++] = 0;
	    }
	}

	++values[i];
	return true;
    }
}
