package toools.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.NetworkInterface;
import java.net.Socket;
import java.net.SocketException;
import java.net.URL;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import toools.CodeShouldNotHaveBeenReachedException;
import toools.collections.Collections;
import toools.io.Utilities;
import toools.math.relation.HashRelation;
import toools.math.relation.Relation;
import toools.thread.CollectionMultiThreadProcessing;
import toools.thread.OneElementOneThreadProcessing;


public class NetUtilities
{
    public static byte[] retrieveURLContent(URL url) throws IOException
    {
	InputStream is = url.openConnection().getInputStream();
	byte[] bytes = toools.io.Utilities.readUntilEOF(is);
	is.close();
	return bytes;
    }

    public List<InetAddress> getInetAddresses(List<String> hostnames) throws UnknownHostException
    {
	List<InetAddress> l = new ArrayList<InetAddress>();

	for (String s : hostnames)
	{
	    l.add(InetAddress.getByName(s));
	}

	return l;
    }

    
    public static String retrieveURLContentAsString(String url) throws IOException
    {
	return new String(retrieveURLContent(new URL(url)));
    }

    public static byte[] retrieveURLContent(String url) throws IOException
    {
	return retrieveURLContent(new URL(url));
    }

    public static byte[] retrieveURLContent(String cgiAddress, byte[] postData)
    throws IOException
    {
	return retrieveURLContent(cgiAddress, new HashMap(), postData);
    }

    
    public static byte[] retrieveURLContent(String cgiAddress, Map<String, String> parms, byte[] postData)
	    throws IOException
    {
	String rq = "";

	if (!parms.isEmpty())
	{
	    for (String key : parms.keySet())
	    {
		key = URLEncoder.encode(key, "UTF-8");
		String value = URLEncoder.encode(parms.get(key), "UTF-8");

		rq += (rq.isEmpty() ? "?" : "&") + key + "=" + value;
	    }
	}

	URL url = new URL(cgiAddress + rq);
	java.net.URLConnection conn = url.openConnection();

	if (postData != null)
	{
	    conn.setDoOutput(true);
	    OutputStream os = conn.getOutputStream();
	    os.write(postData);
	    os.flush();
	}

	// System.out.println(conn.getHeaderFields());
	InputStream is = conn.getInputStream();
	byte[] response = Utilities.readUntilEOF(is);
	is.close();
	return response;
    }

    public static String determineLocalHostName()
    {
	List<String> names = new ArrayList<String>();

	try
	{
	    names.add(InetAddress.getLocalHost().getCanonicalHostName());
	}
	catch (UnknownHostException e)
	{
	}

	try
	{
	    names.add(InetAddress.getLocalHost().getHostName());
	}
	catch (UnknownHostException e)
	{
	}

	try
	{
	    names.add(InetAddress.getLocalHost().getHostAddress());
	}
	catch (UnknownHostException e)
	{
	}

	for (String name : names)
	{
	    if (name != null)
	    {
		return name;
	    }
	}

	throw new CodeShouldNotHaveBeenReachedException();
    }

    public static byte[] retrieveLucHogieData(String name)
    {
	try
	{
	    return retrieveURLContent(new URL("http://luc.hogie.fr/" + name));
	}
	catch (MalformedURLException e)
	{
	    e.printStackTrace();
	    return null;
	}
	catch (IOException e)
	{
	    e.printStackTrace();
	    return null;
	}
    }

    public static Socket connect(InetAddress ip, int port, int timeoutms) throws IOException
    {
	Socket s = new Socket();

	try
	{
	    s.connect(new InetSocketAddress(ip, port), timeoutms);

	}
	catch (ConnectException e)
	{
	    ConnectException c = new ConnectException(e.getMessage() + " to " + ip.getHostName() + " on port " + port);
	    c.setStackTrace(e.getStackTrace());
	    throw c;
	}

	return s;
    }

    public static boolean isServerRunningOnPort(InetAddress host, int port, int timeoutms)
    {
	try
	{
	    Socket s = connect(host, port, timeoutms);

	    try
	    {
		s.close();
	    }
	    catch (Throwable t)
	    {
	    }

	    return true;
	}
	catch (IOException e)
	{
	    return false;
	}
    }

    public static boolean isLocalServerRunningOnPort(int port, int timeoutms)
    {
	try
	{
	    return isServerRunningOnPort(InetAddress.getLocalHost(), port, timeoutms);
	}
	catch (UnknownHostException e)
	{
	    e.printStackTrace();
	    return false;
	}
    }

    public static boolean isLocalSMTPServerRunning()
    {
	return isLocalServerRunningOnPort(25, 1000);
    }

    public static void sendEmailUsingLocalSMTPServer(String sender, String recipient, String subject, String content)
	    throws MessagingException
    {
	System.setProperty("mail.smtp.starttls.enable", "true");
	System.setProperty("mail.smtp.host", "localhost");
	Session session = Session.getDefaultInstance(System.getProperties(), null);
	MimeMessage message = new MimeMessage(session);
	message.setFrom(new InternetAddress(System.getProperties().getProperty("user.name")));
	message.addRecipient(Message.RecipientType.TO, new InternetAddress(recipient));
	message.setSubject(subject);
	message.setText(content);
	Transport.send(message);
    }

    public static void sendEmail(String sender, String recipient, String subject, String content, String smtpServer,
	    String username, String passwd) throws MessagingException
    {
	System.setProperty("mail.smtp.starttls.enable", "true");
	Session session = Session.getDefaultInstance(System.getProperties(), null);
	MimeMessage message = new MimeMessage(session);
	message.setFrom(new InternetAddress(sender));
	message.addRecipient(Message.RecipientType.TO, new InternetAddress(recipient));
	message.setSubject(subject);
	message.setText(content);
	Transport tr = session.getTransport("smtp");
	tr.connect(smtpServer, username, passwd);
	message.saveChanges(); // don't forget this
	tr.sendMessage(message, message.getAllRecipients());
	tr.close();

    }

    public static Relation<NetworkInterface, InetAddress> getNonLoopbackIPv4LocalAddresses()
    {
	Relation<NetworkInterface, InetAddress> ips = new HashRelation<NetworkInterface, InetAddress>();
	Relation<NetworkInterface, InetAddress> aaa = getIPAddresses();

	for (NetworkInterface ni : aaa.getKeys())
	{
	    for (InetAddress localIP : aaa.getValues(ni))
	    {
		// if this is a non-loopback IPv4 address
		if (!localIP.isLoopbackAddress() && localIP.getAddress().length == 4)
		{
		    ips.add(ni, localIP);
		}
	    }
	}

	return ips;
    }

    public static Collection<Socket> getWorkingServersInLAN(int port, int timeout)
    {
	return getWorkingServersInLANOf(port, getLocalAddress(), timeout);
    }

    public static Collection<Socket> getWorkingServersInLANOf(int port, InetAddress localAddress, int timeout)
    {
	return getWorkingServers(port, getAllIPAddressesInLAN(localAddress), timeout);
    }

    public static Collection<Socket> getWorkingServers(final int port, Collection<InetAddress> candidateServers,
	    final int timeoutms)
    {
	final Collection<Socket> servers = new ArrayList<Socket>();
	final List<InetAddress> l = new ArrayList<InetAddress>(candidateServers);

	new OneElementOneThreadProcessing(candidateServers) {
	    @Override
	    protected Object run(List<Thread> threads, int threadIndex, int index)
	    {
		InetAddress canditate = l.get(index);

		try
		{
		    servers.add(connect(canditate, port, timeoutms));
		}
		catch (IOException e)
		{
		}

		return null;
	    }

	};

	return servers;
    }

    public static InetAddress getLocalAddress()
    {
	try
	{
	    return InetAddress.getLocalHost();
	}
	catch (UnknownHostException e)
	{
	    e.printStackTrace();
	    throw new IllegalStateException();
	}
    }

    public static void main(String[] args) throws IOException
    {
	for (Socket s : getWorkingServersInLAN(62592, 1000))
	{
	    System.out.println(s.getRemoteSocketAddress());
	}
    }

    public static void closeAll(Collection<Socket> sockets)
    {
	for (Socket s : sockets)
	{
	    try
	    {
		s.close();
	    }
	    catch (IOException t)
	    {
	    }
	}
    }

    public static Collection<NetworkInterface> getNetworkInterfaces()
    {
	try
	{
	    return Collections.convertEnumerationToCollection(NetworkInterface.getNetworkInterfaces());
	}
	catch (SocketException ex)
	{
	    ex.printStackTrace();
	    throw new IllegalStateException();
	}
    }

    public static NetworkInterface getNetworkInterface(String name)
    {
	for (NetworkInterface ni : getNetworkInterfaces())
	{
	    if (ni.getDisplayName().equals(name))
	    {
		return ni;
	    }
	}

	return null;
    }

    public static Relation<NetworkInterface, InetAddress> getIPAddresses()
    {
	Relation<NetworkInterface, InetAddress> addresses = new HashRelation<NetworkInterface, InetAddress>();

	for (NetworkInterface netInterface : getNetworkInterfaces())
	{
	    addresses.addAll(netInterface, Collections.convertEnumerationToCollection(netInterface.getInetAddresses()));
	}

	return addresses;
    }

    public static Collection<InetAddress> getAllIPAddressesInLAN(InetAddress ip)
    {
	if (ip.isLoopbackAddress())
	    throw new IllegalArgumentException(ip + " is a loopback address");

	if (ip.getAddress().length != 4)
	    throw new IllegalArgumentException(ip + " is not an IPv4 address");

	Collection<InetAddress> addresses = new ArrayList<InetAddress>();

	byte[] bytes = ip.getAddress();

	for (int i = 1; i < 255; ++i)
	{
	    bytes[3] = (byte) i;

	    try
	    {
		addresses.add(InetAddress.getByAddress(bytes));
	    }
	    catch (UnknownHostException e)
	    {
		e.printStackTrace();
	    }
	}

	return addresses;
    }

    public static Collection<? extends InetAddress> getAllIPAddressesInLAN()
    {
	return getAllIPAddressesInLAN(getLocalAddress());
    }

    public static boolean isURL(String sourceFileName)
    {
	try
	{
	    new URL(sourceFileName);
	    return true;
	}
	catch (MalformedURLException e)
	{
	    return false;
	}
    }


}
